#! /usr/bin/env python
# coding: utf-8

import glob
import os
import numpy as np
import pandas
import xarray

from smartcarb import misc
from smartcarb import io

import smartcarb.config as cfg
import dplume


def s5_to_s7_filename(fn):
    """
    Get Sentinel-7 filename from plume detection using Sentinel-5.
    Sentinel-7 filename is part of the filename.
    """
    sat, orbit, time, lon_eq = io.process_filename(fn)
    fn = io.create_filename(time, input_files='level2', instrument=sat,
                            orbit=orbit, lon_eq=lon_eq)
    return fn



def read_s7_for_s5_result(filename, origin='Berlin'):
    """
    Reads a Sentinel-7 that is close to Sentinel-5 orbit used
    by plume detection.
    """
    fn7 = s5_to_s7_filename(filename)
    data7 = io.read_level2(fn7, co2_noise=0.5, co2_thr=0.01,
                           no2_noise='high', no2_thr=0.30, origin=origin)
    data7['CO2_true_plume'] = dplume.create_true_plume(data7, 'CO2', origin,
                                                       0.05, 1.0)
    return data7



def read_overpasses(origin, nsat=None):

    path = "/project/kug/smartcarb/plume_detection/"
    filename = os.path.join(path, origin, "%sPlume.dat" % origin)
    data = pandas.read_csv(filename, sep='\t', index_col=2, parse_dates=True)

    if nsat is not None:
        data = misc.filter_dataset(data, nsat=nsat)

    return data



def get_product(tracer, ssp, q, ns, use_s5=False, use_mc=True):
    if use_s5:
        ssp = 'S5'

    ns = ('%03d' if use_mc else '%d') % ns

    if tracer == 'CO2':
        product = '%s_%.1f_q%.3f_ns%s' % (tracer, ssp, q, ns)
    else:
        product = '%s_%s_q%.3f_ns%s' % (tracer, ssp, q, ns)

    return product


def get_path_to_results(origin, use_mc=True):
    path = '/project/kug/smartcarb/results'
    folder = 'plume_detection_mc' if use_mc else 'plume_detection'
    path = os.path.join(path, origin, folder)
    return path


def average_results(ts, add_size=False):
    """
    Mean and standard deviation of time series by combining
    values with identical index, orbits and lon_eqs.
    """
    indices = []
    means = []
    stds = []
    size = []

    for index, orbit, lon_eq in sorted(set(zip(ts.index, ts.orbit.values,
                                               ts.lon_eq.values))):
        mask  = (ts.index == index)
        mask &= (ts.orbit == orbit)
        mask &= (ts.lon_eq == lon_eq)

        indices.append(index)
        means.append(ts[mask].mean())
        stds.append(ts[mask].std())
        size.append(mask.sum())

    if add_size:
        return pandas.DataFrame(means, indices), pandas.DataFrame(stds,indices), \
               pandas.DataFrame(size, indices, columns=["nsamples"])
    else:
        return pandas.DataFrame(means, indices), pandas.DataFrame(stds, indices)


def read_true_plumes(nsat=None):
    d = pandas.read_csv("/nas/smartcarb/output/timeseries/CO2_Berlin_plume_size.dat",
                        delimiter='\t', index_col=0, parse_dates=True,
                        comment='#')

    if nsat is not None:
        d = misc.filter_dataset(d, nsat=nsat)

    return d

def filter_overlap(ts, nmax=100):
    """
    Remove satellites and orbits which don't overlap Berlin and have less than
    nmax real pixels.
    """
    d = read_true_plumes()
    does_overlap = np.array(d.does_overlap).flatten()
    n_pixels = np.array(d['pixels (<1%)']).flatten()
    mask = []

    for index, col in ts.iterrows():
        m = (d.index == index) & (d.orbit == col.orbit) & (d.lon_eq == col.lon_eq)

        if np.sum(m) == 1:
            m = does_overlap[m] & (n_pixels[m] >= nmax)
        else:
            m = False

        mask.append(m)

    return ts[np.array(mask)]



def read_results(tracer, ssp, ns, q=0.99, nsat=6, origin='Berlin', nmax=None,
                 use_mc=True):

    path = get_path_to_results(origin, use_mc=use_mc)

    product = get_product(tracer, ssp, q, ns, use_mc=use_mc)
    filename = os.path.join(path, '%s.dat' % product)

    # read and filter by constellation
    data = pandas.read_csv(filename, sep='\t', index_col=0, parse_dates=True,
                           comment='#')
    data = misc.filter_dataset(data, nsat=nsat)

    # remove plumes with less than nmax CO2 pixels
    if nmax is not None:
        if origin == 'Berlin' and 'TP1' in data:
            data = data[data.TP1 >= nmax]
        else:
            data = data[data.CO2_pixels >= nmax]

    # add sample number (assuming 20 samples)
    if use_mc:
        data['nsample'] = np.tile(np.arange(20, dtype=int), data.shape[0]//20)

        data['filename'] = [get_output_filename('Sentinel_7_CO2', v.orbit, v.lon_eq, time,
                                                tracer, ssp, ns, origin,
                                                sample=v.nsample, use_mc=use_mc)
                            for time, v in data.iterrows()]

    # add filenames
    else:
        data['filename'] = [get_output_filename(v.satellite, v.orbit, v.lon_eq, time,
                                                tracer, ssp, ns, origin)
                            for time, v in data.iterrows()]

    return data


def output2input_filename(filename):
    sat, orbit, time, lon_eq = io.process_filename(filename)
    return io.create_filename(time, input_files='level2',
                              instrument=sat, orbit=orbit,
                              lon_eq=lon_eq)



def read_output(filename, ssp=None):
    data = xarray.open_dataset(filename)
    data.attrs['time'] = pandas.Timestamp(data.time)

    # replace CO2 and CO2 noise field with different noise scenario 
    if ssp is not None:
        if data.attrs['tracer'] == 'CO2':
            print("""\
Warning: CO2 will be overwritten but CO2 was used for plume detection.
         Derived data arrays might be inconsistent."""
            )

        l2filename = output2input_filename(filename)
        level2 = io.read_level2(l2filename, co2_noise=ssp, co2_thr=0.01)

        data['CO2'] = level2['CO2']
        data['CO2_noise'] = level2['CO2_noise']


    # add noise level (if it's missing)
    for noise in [0.5, 0.7, 1.0]:
        if '%.1f' % noise in data.CO2_noise.attrs['standard name'] \
           and 'noise_level' not in data.CO2.attrs:
            data.CO2.attrs['noise_level'] = noise

    return data


def get_sample_filenames(time, tracer, ssp, ns=37, orbit=None, lon_eq=None,
                         nsample=0, origin='Berlin'):
    """
    Get the sample filenames matching the description.

    TODO: Sort by lon_eq i.e. number of satellites.
    """

    if nsample is None:
        nsample = '??'

    use_mc = nsample is not None
    product = get_product(tracer, ssp, q=0.99, ns=ns, use_mc=use_mc)

    filename = '_'.join([
        'Sentinel_7_CO2',
        time.strftime('%Y%m%d%H'),
        'o????' if orbit is None else 'o%04d' % orbit,
        'l????' if lon_eq is None else 'l%04d' % lon_eq,
        product.replace('0', '*'),
        'n%s.nc' % nsample if isinstance(nsample, str) else 'n%02d.nc' % nsample,
    ])

    filename = os.path.join(get_path_to_results(origin, use_mc), product, filename)
    print(filename)

    return sorted(glob.glob(filename))



def get_output_filename(sat, orbit, lon_eq, time, tracer, ssp, ns, origin,
                        sample=None, use_mc=False):

    product = get_product(tracer, ssp, q=0.99, ns=ns)

    if use_mc:
        path = '/project/kug/smartcarb/results/%s/plume_detection_mc' % origin
    else:
        path = '/project/kug/smartcarb/results/%s/plume_detection' % origin

    time = time.strftime('%Y%m%d%H')

    if sample is None:
        filename = '%s_%s_o%04d_l%04d_%s.nc' % (sat, time, orbit, lon_eq, product)
    else:
        filename = '%s_%s_o%04d_l%04d_%s_n%02d.nc' % (sat, time, orbit, lon_eq,
                                                      product, sample)

    filename = os.path.join(path, product, filename)

    return filename


def iter_level2(data, tracer, ssp, ns, q=0.99, origin='Berlin'):
    for time, v in data.iterrows():
        filename = get_output_filename(v.satellite, v.orbit, v.lon_eq, time, tracer,
                                       ssp, ns, origin)

        d = read_output()

        yield d








