
import shapely
import numpy as np

import smartcarb as sc


def overlaps_with_area(lon, lat, area):
    points = np.transpose([lon, lat])
    points = shapely.geometry.MultiPoint(points)
    return points.intersects(area)


def get_distances(data, delta, min_distance=-30e3):
    dmax = np.nanmax(data.xp.values[data.detected_plume])
    return np.arange(min_distance, dmax+delta, delta)


def overlaps_with_source(data, source=None, lon_o=None, lat_o=None, radius=None, mask=None):

    if source is not None:
        lon_o, lat_o, default_radius = sc.misc.get_source_origin(source)

        if radius is None:
            radius = default_radius

    valids = np.isfinite(data.lon) & np.isfinite(data.lat)
    
    if mask is not None:
        valids = valids & ~mask
            
    lon = np.array(data.lon)[valids]
    lat = np.array(data.lat)[valids]

    area = sc.misc.get_circle(lon_o, lat_o, radius)

    return overlaps_with_area(lon, lat, area)


def compute_snrs(signals, ns=1.0, bg_means=0.0, bg_stds=0.0):
    """
    Compute signal-to-noise ratio from satellite observation.

    signals :: signal strength (CO2, NO2 and CO)
    bg_stds :: standard deviation of background (CO2, NO2 and CO)
    ns      :: size of neighborhood used for computing signal
    """
    if np.ndim(bg_means) == 0:
        bg_means = np.full_like(signals, bg_means)

    if np.ndim(bg_stds) == 0:
        bg_stds = np.full_like(signals, bg_stds)

    snrs = []
    for i,tracer in enumerate(['CO2', 'NO2', 'CO']):
        if tracer == 'CO2':
            ssp = np.array([0.5, 0.7, 1.0])

        elif tracer == 'NO2':
            no2 = signals[i] + bg_means[i]
            ssp = np.max([[0.1, 0.2],
                          [no2 * 0.15, no2 * 0.20]], axis=0)
        elif tracer == 'CO':
            ssp = np.array([4.0]) # 10^17

        snr = signals[i] / np.sqrt(ssp**2 / ns + bg_stds[i]**2)
        snrs.append(snr)

    return np.concatenate(snrs)



